#ifdef __FreeBSD__
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/bpf.h>

#include "pktlib.h"
#include "lib.h"

struct _pktif {
  struct _pktif_base base; /* ΰƬ֤ɬפ */
  struct { /* 󥿡եBPF¸ʬ */
    char *buffer;
    int bufsize;
    int readsize;
    struct bpf_hdr *hdr;
  } bpf;
};

pktif_t pktif_open(char *ifname, unsigned long flags, int option_size)
{
  pktif_t pktif;
  int size, fd, buffer_size;
  struct ifreq ifr;
  unsigned int one = 1, val;

  size = sizeof(*pktif) + option_size + strlen(ifname) + 1;
  pktif = malloc(size);
  memset(pktif, 0, size);

  fd = open("/dev/bpf", O_RDWR); /* BPF򳫤 */
  if (fd < 0)
    pktlib_error_exit("Cannot open bpf.\n");

  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
  if (ioctl(fd, BIOCSETIF, &ifr) < 0) /* 󥿡եꤹ */
    pktlib_error_exit("Fail to ioctl BIOCSETIF.\n");

  /* μ */
  if (ioctl(fd, BIOCGBLEN, &val) < 0) /* Хåեɬץ */
    pktlib_error_exit("Fail to ioctl BIOCGBLEN.\n");
  buffer_size = val;

  /* Ϣ */
  if (!(flags & PKTIF_OPEN_FLAG_RECV_NOTPROM)) {
    if (ioctl(fd, BIOCPROMISC, NULL) < 0) /* ǤʤѥåȤ */
      pktlib_error_exit("Fail to ioctl BIOCPROMISC.\n");
  }
  if (ioctl(fd, BIOCIMMEDIATE, &one) < 0) /* ¨read() */
    pktlib_error_exit("Fail to ioctl BIOCIMMEDIATE.\n");
  val = (flags & PKTIF_OPEN_FLAG_RECV_NOTSENT) ? 0 : 1;
  if (ioctl(fd, BIOCSSEESENT, &val) < 0) /* ϥѥåȤ */
    pktlib_error_exit("Fail to ioctl BIOCSSEESENT.\n");
  if (ioctl(fd, BIOCFLUSH, NULL) < 0) /* Хåեեå夹 */
    pktlib_error_exit("Fail to ioctl BIOCFLUSH.\n");

  /* Ϣ */
  if (ioctl(fd, BIOCSHDRCMPLT, &one) < 0) /* MACɥ쥹֤ʤ */
    pktlib_error_exit("Fail to ioctl BIOCSHDRCMPLT.\n");

  pktif->base.next = NULL;
  pktif->base.name = (char *)(pktif + 1) + option_size;
  strcpy(pktif->base.name, ifname);
  pktif->base.fd = fd;
  pktif->base.flags = flags;
  pktif->base.buffer_size = buffer_size;
  pktif->base.option = (option_size > 0) ? (pktif + 1) : NULL;
  pktif->bpf.buffer = malloc(buffer_size); /* Хåե */
  pktif->bpf.bufsize = buffer_size;
  pktif->bpf.readsize = 0;

  pktlib_iflist_set(pktif); /* select()Τ˥󥯥ꥹȤ³ */

  return pktif;
}

int pktif_is_empty(pktif_t pktif)
{
  return (pktif->bpf.readsize == 0) ? 1 : 0;
}

int pktif_recv(pktif_t pktif, char *buffer, int size, struct timeval *tm)
{
  struct bpf_hdr *hdr;
  int r;

  /*
   * BPFǤϥإåղäʣѥåȤ֤Τǡread()
   * ȤμѥåȤХåե˻ĤäƤˤϤ֤
   * ĤäƤʤʤread()ǿ˼롥
   */

  if (pktif_is_empty(pktif)) { /* ѥåȤμ */
    r = read(pktif->base.fd, pktif->bpf.buffer, pktif->bpf.bufsize);
    if (r <= 0)
      return r;
    pktif->bpf.readsize = r;
    pktif->bpf.hdr = (struct bpf_hdr *)pktif->bpf.buffer;
  }

  hdr = pktif->bpf.hdr;

  if (tm) {
    tm->tv_sec  = hdr->bh_tstamp.tv_sec;
    tm->tv_usec = hdr->bh_tstamp.tv_usec;
  }
  if (size > hdr->bh_caplen)
    size = hdr->bh_caplen;
  memcpy(buffer, (char *)hdr + hdr->bh_hdrlen, hdr->bh_caplen);

  pktif->bpf.hdr = (struct bpf_hdr *)
    ((char *)hdr + BPF_WORDALIGN(hdr->bh_hdrlen + hdr->bh_caplen));

  if ((char *)pktif->bpf.hdr >= pktif->bpf.buffer + pktif->bpf.readsize)
    pktif->bpf.readsize = 0;

  return size;
}

int pktif_send(pktif_t pktif, char *buffer, int size)
{
  return write(pktif->base.fd, buffer, size); /* ѥåȤ */
}
#endif
